import numpy as np
import itertools


from scipy.sparse import csr_matrix, tril, kron
from scipy.linalg import block_diag
from itertools import combinations
from functools import lru_cache
from numpy.linalg import inv, cholesky
from .utils import (
    mc_sandwich,
    diagonal,
    mc_sandwich_power,
    mc_matrix_linear_predictor,
    mc_sandwich_power_csr,
)


class MCGLMCAttributes:
    """
    The class "MCGLMCAttributes" has the sake of calculating every C operations, used on throughout adjustments of mean and variance. This class has two interfaces, "c_inverse" and "c_complete"; one for each of two adjustment steps of MCGLM.

    The interface "c_inverse" crafts only inverse C, and the "c_complete" adds its derivatives and other features onto response. A Quasi-likelihood estimation needs only the inverse of "C" matrix. Therefore c_inverse saves computational resources by avoiding unnecessary operations on mean step adjustment.
    """

    def c_inverse(self, mu, power, rho, tau, full_response=False):
        """
        A method to generate only the inverse of the C matrix, explicitly made for the mean treatment step. This method interacts with sigma and omega amenities by list of each parameter.

        Parameters
        ----------
            mu : array_like
                A vetor with mean parameters.
            power : float
                Power parameter.
            rho : float
                Correlation parameter.
            tau : float
                Dispersion parameter.
        Returns
        -------
            array_like or tuple : The inverse of C matrix and its components.
        """
        c_inverse = self.__generate_c_inverse(mu, power, rho, tau, full_response)
        return c_inverse

    def c_complete(self, mu, power, rho, tau):
        """
        A method to generate the whole list of C components, explicitly made for the variance treatment step. This method interacts with sigma and omega crafting practices, passing the list of each parameter.

        Parameters
        ----------
            mu : array_like
                A vetor with mean parameters.
            power : float
                Power parameter.
            rho : float
                Correlation parameter.
            tau : float
                Dispersion parameter.
        Returns
        -------
            tuple : A tuple with every component of C.
        """
        (
            diagonal_matrix,
            omega,
            sigma_raw,
            sigma_chol,
            sigma_chol_inv,
            sigma_between,
            sigma_between_derivative,
            sigma_chol_block_matrix,
            sigma_chol_inv_block_matrix,
            c_inverse,
        ) = self.__generate_c_inverse(mu, power, rho, tau, full_response=True)

        sigma_derivatives = self.__generate_sigma_derivatives(omega, mu, power)

        core_matrix = np.kron(sigma_between, diagonal_matrix)
        sigma_chol_block_matrix_transpose = sigma_chol_block_matrix.transpose()
        c_derivatives = self.__generate_derivative_c(
            sigma_chol,
            sigma_chol_inv,
            sigma_derivatives,
            core_matrix,
            sigma_chol_block_matrix,
            sigma_chol_block_matrix_transpose,
            sigma_between_derivative,
            diagonal_matrix,
            self._n_targets,
        )
        c_values = self.__generate_c_values(
            sigma_between,
            diagonal_matrix,
            sigma_chol_block_matrix_transpose,
            sigma_chol_block_matrix,
        )

        return c_inverse, c_derivatives, c_values

    def __generate_c_inverse(self, mu, power, rho, tau, full_response=False):
        """This method generates and retrieves the inverse of the C matrix, which is a pivotal component of the quasi-score calculation. Notwithstanding the mentioned, this method crafts many important artifacts throughout the time as omega, sigma, sigma Cholesky, the inverse of Cholesky sigma, sigma among responses, sigma block diagonal matrix alongside its inverse.
        Owing to the previously mentioned worthy artifacts crafted by the method, it can retrieve either only the c inverse matrix or all artifacts. The former is helpful for the method _c_inverse, whereas the former is for _c_complete.

        Parameters
        ----------
            mu : array-like
                A vetor with mean parameters.
            power : float
                Power parameter.
            rho : float
                Correlation parameter.
            tau : float
                Dispersion parameter.
        Returns
        -------
            array_like : A matrix with the inverse matrix of C.
        """
        diagonal_matrix = diagonal(self._n_obs, np.ones(self._n_obs))

        omega = self._generate_omega(tau)
        build_sigma = map(
            self._calculate_sigma,
            mu,
            power,
            omega,
            self._variance,
            self._ntrial,
        )
        sigma_raw, sigma_chol, sigma_chol_inv = self.__parser_sigma(build_sigma)

        sigma_between, sigma_between_derivative = self._sigma_between_values(
            rho=rho, n_resp=self._n_targets
        )
        raw_c_components = self.__generate_c_inverse_and_blocks(
            sigma_chol, sigma_chol_inv, sigma_between, diagonal_matrix
        )

        (
            sigma_chol_block_matrix,
            sigma_chol_inv_block_matrix,
            c_inverse,
        ) = self.__parser_c_inverse(raw_c_components)

        if not full_response:
            return c_inverse
        else:
            return (
                diagonal_matrix,
                omega,
                sigma_raw,
                sigma_chol,
                sigma_chol_inv,
                sigma_between,
                sigma_between_derivative,
                sigma_chol_block_matrix,
                sigma_chol_inv_block_matrix,
                c_inverse,
            )

    def __generate_sigma_derivatives(self, omega, mu, power):
        """Base method to generate derivatives related to sigma.

        Parameters
        ----------
            omega : array-like
                The omega matrix.
            mu : array-like
                A vetor with mean parameters.
            power : float
                The power parameter.
        Returns
        -------
            _type_: _description_
        """
        build_sigma = map(
            self._calculate_sigma_derivatives,
            mu,
            power,
            self._variance,
            self._z,
            self._power_fixed,
            self._ntrial,
            omega,
        )

        sigma_derivative = self.__parser_sigma_derivatives(build_sigma)
        return sigma_derivative

    def __generate_derivative_c(
        self,
        sigma_chol,
        sigma_chol_inv,
        sigma_derivatives,
        core_matrix,
        sigma_chol_block_matrix,
        sigma_chol_block_matrix_transpose,
        sigma_between_derivative,
        diagonal_matrix,
        n_target,
    ):
        """Base method to generate derivatives related to C matrix.

        Parameters
        ----------
            sigma_chol : array_like
                Sigma matrix decomposed by the Cholesky operation.
            sigma_chol_inv : array_like
                Inverse sigma matrix decomposed by the Cholesky operation.
            sigma_derivatives : array_like
                Derivatives related to the Sigma.
            core_matrix : array_like
                An intermediate matrix for Sigma operation
            sigma_chol_block_matrix : array_like
                A block diagonal matrix with all Sigmas.
            sigma_chol_block_matrix_transpose : array_like
                The transpose matrix of a block diagonal matrix with all Sigmas.
            sigma_between_derivative : array_like
                The deriviatives between sigmas.
            diagonal_matrix : array_like
                A diagonal matrix.
            n_target : int
                Total outcome variables.
        Returns
        -------
            array_like : A matrix with derivatives related to the C Matrix.
        """
        if n_target == 1:

            core_matrix_csr = sigma_between_derivative
            sigma_chol_block_matrix_csr = sigma_chol_block_matrix
            sigma_chol_block_matrix_transpose_csr = csr_matrix(
                sigma_chol_block_matrix_transpose
            )
            sigma_between_derivative_csr = sigma_between_derivative

        else:
            core_matrix_csr = csr_matrix(core_matrix)
            sigma_chol_block_matrix_csr = csr_matrix(sigma_chol_block_matrix)
            sigma_chol_block_matrix_transpose_csr = csr_matrix(
                sigma_chol_block_matrix_transpose
            )
            sigma_between_derivative_csr = []
            for sigma in sigma_between_derivative:
                sigma_between_derivative_csr.append(csr_matrix(sigma))

            diagonal_matrix_csr = csr_matrix(diagonal_matrix)

        list_d_chol_sigma = list(
            map(
                self.__derivative_cholesky,
                sigma_derivatives,
                sigma_chol_inv,
                sigma_chol,
            )
        )
        bdiag_d_chol_sigma = list(
            map(
                self.__mc_transform_list_bdiag,
                list_d_chol_sigma,
                list(range(n_target)),
                [n_target for _ in range(n_target + 1)],
            )
        )
        bdiag_d_chol_sigma_csr = []
        for sigmas in bdiag_d_chol_sigma:
            bdiag_d_chol_sigma_csr.append([csr_matrix(sigma) for sigma in sigmas])
        if n_target == 1:
            bdiag_d_chol_sigma_csr = bdiag_d_chol_sigma_csr[0]
            d_c = [
                mc_sandwich_power_csr(
                    core_matrix_csr,
                    d_chol_sigma,
                    sigma_chol_block_matrix_transpose_csr,
                ).toarray()
                for d_chol_sigma in bdiag_d_chol_sigma_csr
            ]
            D_C = d_c
        else:
            bdiag_d_chol_sigma = list(itertools.chain(*list(bdiag_d_chol_sigma_csr)))

            d_c = [
                mc_sandwich_power_csr(
                    core_matrix_csr,
                    d_chol_sigma,
                    sigma_chol_block_matrix_transpose_csr,
                ).toarray()
                for d_chol_sigma in bdiag_d_chol_sigma
            ]
            d_c_rho = self.__derivative_rho(
                sigma_between_derivative_csr,
                sigma_chol_block_matrix_csr,
                sigma_chol_block_matrix_transpose_csr,
                diagonal_matrix_csr,
            )
            D_C = d_c_rho + d_c
        return D_C

    def __generate_c_values(
        self,
        sigma_between,
        diagonal_matrix,
        sigma_chol_block_matrix_transpose,
        sigma_chol_block_matrix,
    ):
        """base method to generate the C matrix.

        Parameters
        ----------
            sigma_between : array_like
                Matrix with inner values among sigmas.
            diagonal_matrix : array_like
                A diagonal matrix.
            sigma_chol_block_matrix_transpose : array_like
                The transpose matrix of a block diagonal matrix with all Sigmas.
            sigma_chol_block_matrix : array_like
                A block diagonal matrix with all Sigmas.
        Returns
        -------
            array_like : A matrix with derivatives related to the C Matrix.
        """
        kron_middle = np.kron(sigma_between, diagonal_matrix)

        c_computation = np.dot(
            np.dot(
                sigma_chol_block_matrix_transpose,
                kron_middle,
            ),
            sigma_chol_block_matrix,
        )
        return c_computation

    def __derivative_rho(
        self,
        d_sigmab,
        sigma_chol_block_matrix,
        sigma_chol_block_matrix_transpose,
        core_matrix,
    ):
        """Base method to calculate derivatives related to correlation parameters.

        Parameters
        ----------
            d_sigmab : array_like
                Matrix with inner values among sigmas.
            sigma_chol_block_matrix : array_like
                A block diagonal matrix with all Sigmas.
            sigma_chol_block_matrix_transpose : array_like
                The transpose matrix of a block diagonal matrix with all Sigmas.
            core_matrix : array_like
                A diagonal matrix.
        Returns
        -------
            array_like : A matrix with derivatives related to the C Matrix.
        """

        def multiplication_matrix(
            sigmab,
            sigma_chol_block_matrix,
            sigma_chol_block_matrix_transpose,
            core_matrix,
        ):
            kron_matrix = kron(sigmab, core_matrix)
            return (sigma_chol_block_matrix_transpose.dot(kron_matrix)).dot(
                sigma_chol_block_matrix
            )

        return [
            multiplication_matrix(
                sigmab,
                sigma_chol_block_matrix,
                sigma_chol_block_matrix_transpose,
                core_matrix,
            ).toarray()
            for sigmab in d_sigmab
        ]

    def _generate_omega(self, tau):
        """Base method to calculate derivatives related to correlation parameters.

        Parameters
        ----------
            tau : list
                List with dispersion parameters
        Returns
        -------
            list : the results of matrix linear sum between tau and dependence matrices.
        """
        omega = []
        for target in range(self._n_targets):
            omega.append(mc_matrix_linear_predictor(tau=tau[target], z=self._z[target]))

        return omega

    def __generate_c_inverse_and_blocks(
        self,
        sigma_chol,
        sigma_chol_inv,
        sigma_between,
        diagonal_matrix,
    ):
        """Base method to calculate derivatives related to correlation parameters.

        Parameters
        ----------
            sigma_chol : array_like
                The sigma matrix decomposed by Cholesky.
            sigma_chol_inv : array_like
                The inverse of sigma matrix decomposed by Cholesky.
            sigma_between : array_like
                The inner-matrix among sigmas.
            diagonal_matrix : array_like
                A diagonal matrix.
        Returns
        -------
            array_like : A matrix with derivatives related to the C Matrix.
        """
        sigma_chol_block_matrix = block_diag(*sigma_chol)
        sigma_chol_inv_block_matrix = block_diag(*sigma_chol_inv)

        try:
            ls_sigma_diagonal = inv(sigma_between)
        except Exception as e:
            ls_sigma_diagonal = np.array([[1]])

        C_inverse = np.dot(
            np.dot(
                sigma_chol_inv_block_matrix,
                np.kron(ls_sigma_diagonal, diagonal_matrix),
            ),
            sigma_chol_inv_block_matrix.transpose(),
        )

        return (
            sigma_chol_block_matrix,
            sigma_chol_inv_block_matrix,
            C_inverse,
        )

    def _calculate_sigma(
        self, mu, power, omega, variance, Ntrial, covariance="identity"
    ):
        """Base method to calculate sigma

        Parameters
        ----------
            mu : array_like
                A vector with expected values.
            power : float
                A power parameter.
            omega : array_like
                The omega resulted matrix.
            variance : str
                The variance function.
            Ntrial : int
                The number of trial. Parameter for Binomial distribution.
        Returns
        -------
            array_like : A matrix with Sigma values.
        """
        if isinstance(variance, list):
            variance = variance[0]

        if variance == "constant":
            sigma_raw = omega
            sigma_chol = cholesky(sigma_raw).T
            sigma_chol_inv = inv(sigma_chol)
        elif variance in ["tweedie", "binomialP", "binomialPQ"]:
            if variance == "tweedie":
                variance = "power"

            variance_components = self.__generate_variance(
                variance_type=variance, mu=mu.get("mu"), power=power, Ntrial=Ntrial
            )
            sigma_raw = mc_sandwich(
                omega,
                variance_components.get("variance_sqrt_output"),
                variance_components.get("variance_sqrt_output"),
            )
            sigma_chol = cholesky(sigma_raw).T
            sigma_chol_inv = inv(sigma_chol)
        elif variance in ["poisson_tweedie", "geom_tweedie"]:
            diagonal_value = [mu.get("mu") ** 2, mu.get("mu")][
                variance == "poisson_tweedie"
            ]

            variance_components = self.__generate_variance(
                variance_type="power", mu=mu.get("mu"), power=power, Ntrial=Ntrial
            )
            sigma_raw = diagonal(len(mu.get("mu")), diagonal_value) + np.dot(
                np.dot(variance_components.get("variance_sqrt_output"), omega),
                variance_components.get("variance_sqrt_output"),
            )
            sigma_chol = cholesky(sigma_raw).T
            sigma_chol_inv = inv(sigma_chol)
        return dict(
            sigma_raw=sigma_raw, sigma_chol=sigma_chol, sigma_chol_inv=sigma_chol_inv
        )

    def _calculate_sigma_derivatives(
        self, mu, power, variance, z, power_fixed, Ntrial, omegas, covariance="identity"
    ):
        """
        Base method for computing variance-covariance matrix, based on variance function and omega matrix. This method will implement for cases where covariance is equal to identity, and variance falls in the list:
        ['constant', 'tweedie', 'binomialP', 'binomialPQ', 'power', 'geom_tweedie', 'poisson_tweedie']

        Parameters
        ----------
            mu : array_like
                A vector with expected values.
            power : float
                A power parameter.
            variance : str
                The variance function.
            z : list
                The list with z matrices for dependencies specification.
            power_fixed : boolean
                The specification of power estimation.
            Ntrial : int
                The number of trial. Parameter for Binomial distribution.
            omegas : list
                a list with omegas.
        """
        if isinstance(variance, list):
            variance = variance[0]
        sigma_derivative = None

        if variance == "constant":
            sigma_derivative = z
        elif variance in ["tweedie", "binomialP", "binomialPQ"]:
            variance_type = variance if variance != "tweedie" else "power"

            variance_components = self.__generate_variance(
                variance_type=variance_type, mu=mu.get("mu"), power=power, Ntrial=Ntrial
            )
            sigma_derivative = [
                mc_sandwich(
                    d_omega,
                    variance_components.get("variance_sqrt_output"),
                    variance_components.get("variance_sqrt_output"),
                )
                for d_omega in z
            ]
            if not power_fixed:
                if variance in ["tweedie", "binomialP"]:
 
                    sigma_derivative_power = mc_sandwich_power(
                        omegas,
                        variance_components.get("variance_sqrt_output"),
                        variance_components.get("derivative_variance_sqrt_power"),
                    )
                    sigma_derivative.insert(0, sigma_derivative_power)
                    #sigma_derivative.append(sigma_derivative_power)

                elif variance == "binomialPQ":
                    sigma_derivative_p = mc_sandwich(
                        omegas,
                        variance_components.get("variance_sqrt_output"),
                        variance_components.get("derivative_variance_sqrt_p"),
                    )
                    sigma_derivative_q = mc_sandwich(
                        omegas,
                        variance_components.get("variance_sqrt_output"),
                        variance_components.get("derivative_variance_sqrt_q"),
                    )
                    sigma_derivative = [
                        sigma_derivative,
                        sigma_derivative_p,
                        sigma_derivative_q,
                    ]
        elif variance in ["poisson_tweedie", "geom_tweedie"]:
            variance_components = self.__generate_variance(
                variance_type="power", mu=mu.get("mu"), power=power, Ntrial=Ntrial
            )
            sigma_derivative = [
                mc_sandwich(
                    d_omega,
                    variance_components.get("variance_sqrt_output"),
                    variance_components.get("variance_sqrt_output"),
                )
                for d_omega in z
            ]
            if not power_fixed:
                sigma_derivative_power = mc_sandwich_power(
                    omegas,
                    variance_components.get("variance_sqrt_output"),
                    variance_components.get("derivative_variance_sqrt_power"),
                )
                sigma_derivative.insert(0, sigma_derivative_power)
        return dict(sigma_derivative=sigma_derivative)

    def _sigma_between_values(self, rho, n_resp):
        def forcesymmetric(sigmab):
            sigmab_t = sigmab.transpose()
            sigmabsymmetric = np.matmul(sigmab_t, sigmab_t.transpose())
            diagonal_value = sigmabsymmetric.diagonal()
            if np.max(diagonal_value) == np.min(diagonal_value):
                return sigmabsymmetric
            else:
                sigmabsymmetric = sigmab
                sigmabsymmetric[np.triu_indices(n_resp, k=1)] = rho
                return sigmabsymmetric

        """sigma between method computes between for sequence calculations.

        It responds out with 2-position-tuple with sigma between and its derivative.
        """
        if n_resp == 1:
            return (1, 1)
        else:
            sigmab = diagonal(n_resp, np.full(n_resp, 1))
            sigmab[np.tril_indices(n_resp, k=-1)] = rho
            sigmab = forcesymmetric(sigmab)

            d_sigmab = self._mc_derivative_sigma_between(n_resp=n_resp)
            return (sigmab, d_sigmab)

    @lru_cache(maxsize=64)
    def _mc_derivative_sigma_between(self, n_resp):
        list_derivatives = list()
        position = list(combinations(range(n_resp), 2))
        n_par = int(n_resp * (n_resp - 1) / 2)
        for index in range(n_par):
            derivative = np.zeros((n_resp, n_resp))
            derivative[position[index][0], position[index][1]] = 1
            derivative[position[index][1], position[index][0]] = 1
            list_derivatives.append(derivative)

        return list_derivatives

    def __derivative_cholesky(self, d_sigma, inv_chol, chol):
        def faux(d_sigma, inv_chol, chol, inv_chol_transpose):
            csr_d_sigma_matrix = csr_matrix(d_sigma)
            csr_inv_chol_matrix = csr_matrix(inv_chol)
            csr_inv_chol_transpose_matrix = csr_matrix(inv_chol_transpose)

            matrix_operations = csr_inv_chol_matrix.dot(csr_d_sigma_matrix)
            matrix_operations = matrix_operations.dot(csr_inv_chol_transpose_matrix)

            matrix_operations = tril(matrix_operations)
            matrix_operations = matrix_operations.toarray()

            diagonal_indices = np.diag_indices_from(matrix_operations)
            matrix_operations[diagonal_indices] = (
                matrix_operations[diagonal_indices] / 2
            )
            return np.dot(chol, matrix_operations)

        cholesky_element = []
        inv_chol_transpose = inv_chol.transpose()
        for derivative in d_sigma:
            if not isinstance(derivative, list):
                cholesky_element.append(
                    faux(derivative, inv_chol, chol, inv_chol_transpose)
                )
            else:
                for der in derivative:
                    cholesky_element.append(
                        faux(der, inv_chol, chol, inv_chol_transpose)
                    )

        if isinstance(derivative, list):
            cholesky_element.reverse()
        return cholesky_element

    def __mc_transform_list_bdiag(self, list_mat, response_number, number_of_labels):
        def build_block_diag(value, response_number):
            block_value = block_diag(value)

            matrix_base = np.zeros(
                (
                    block_value.shape[0] * number_of_labels,
                    block_value.shape[1] * number_of_labels,
                )
            )

            lower = block_value.shape[0] * response_number
            upper = block_value.shape[1] + block_value.shape[1] * response_number

            matrix_base[lower:upper, lower:upper] = block_value

            return matrix_base

        output = [build_block_diag(d_chol, response_number) for d_chol in list_mat]
        return output

    def _generate_variance(
        self, variance_type: str, mu: np.array, power: int = 1, Ntrial: list = 1
    ):
        return self.__generate_variance(variance_type, mu, power, Ntrial)

    def __generate_variance(self, variance_type, mu, power=1, Ntrial=1):
        """Method to apply the variance function on the mean values.

        Parameters
        ----------
            variance_type : array_like
                Type of variance
            mu : array_like
                A vector with expected values.
            power : float
                A power parameter.
            Ntrial : int
                The number of trial. Parameter for Binomial distribution.
        Returns
        -------
            array_like : A matrix with Sigma values.
        """
        if variance_type == "power":
            return self.__power_variance(mu, power)
        elif variance_type == "binomialP":
            return self.__binomialp_variance(mu, power, Ntrial)
        elif variance_type == "binomialPQ":
            return self.__binomialpq_variance(mu, power, Ntrial)

    def __power_variance(self, mu, power):
        mu_power = mu**power
        sqrt_mu_power = np.sqrt(mu_power)
        n_len = len(mu)

        variance_sqrt_output = diagonal(n=n_len, values=sqrt_mu_power)
        derivative_variance_sqrt_power = diagonal(
            n=n_len, values=((mu_power * np.log(mu)) / (2 * sqrt_mu_power))
        )
        derivative_variance_sqrt_mu = (mu ** (power - 1) * power) / (2 * sqrt_mu_power)

        return dict(
            variance_sqrt_output=variance_sqrt_output,
            derivative_variance_sqrt_power=derivative_variance_sqrt_power,
            derivative_variance_sqrt_mu=derivative_variance_sqrt_mu,
        )

    def __binomialp_variance(self, mu, power, ntrial):
        constant = 1 / ntrial
        mu_power = mu**power
        mu_power1 = (1 - mu) ** power
        mu1mu = constant * (mu_power * mu_power1)
        sqrt_mu1mu = np.sqrt(mu1mu)
        n_len = len(mu)

        variance_sqrt_output = diagonal(n=n_len, values=sqrt_mu1mu)
        derivative_variance_sqrt_power = diagonal(
            n=n_len,
            values=(np.log(1 - mu) * mu1mu + np.log(mu) * mu1mu) / (2 * sqrt_mu1mu),
        )
        derivative_variance_sqrt_mu = (
            constant * (mu_power1 * (mu ** (power - 1)) * power)
            - constant * (((1 - mu) ** (power - 1)) * mu_power * power)
        ) / (2 * sqrt_mu1mu)

        return dict(
            variance_sqrt_output=variance_sqrt_output,
            derivative_variance_sqrt_power=derivative_variance_sqrt_power,
            derivative_variance_sqrt_mu=derivative_variance_sqrt_mu,
        )

    def __binomialpq_variance(self, mu, power, ntrial):
        constant = 1 / ntrial
        p = power[0]
        q = power[1]

        mu_p = mu**p
        mu1_q = (1 - mu) ** q
        mu_p_mu_q = mu_p * mu1_q
        mu1mu = mu_p_mu_q * constant
        sqrt_mu1mu = np.sqrt(mu1mu)
        n_len = len(mu)

        denominator1 = 2 * sqrt_mu1mu
        denominator2 = denominator1 * ntrial

        variance_sqrt_output = diagonal(n=n_len, values=sqrt_mu1mu)

        derivative_variance_sqrt_p = diagonal(
            n=n_len, values=(mu_p_mu_q * np.log(mu)) / denominator2
        )

        derivative_variance_sqrt_q = diagonal(
            n=n_len, values=(mu_p_mu_q * np.log(1 - mu)) / denominator2
        )

        derivative_variance_sqrt_mu = (
            constant * (mu1_q * (mu ** (p - 1)) * p)
            - constant * ((1 - mu) ** (q - 1) * mu_p * q)
        ) / denominator1

        return dict(
            variance_sqrt_output=variance_sqrt_output,
            derivative_variance_sqrt_p=derivative_variance_sqrt_p,
            derivative_variance_sqrt_q=derivative_variance_sqrt_q,
            derivative_variance_sqrt_mu=derivative_variance_sqrt_mu,
        )

    def __parser_sigma(self, build_sigma_map):
        """parser method for _calculate_sigma output attributes
        Args:
            build_sigma_map (list): a list with dicts, with sigmas
        """
        list_sigmas = [
            [
                build_sigma_iteration.get("sigma_raw"),
                build_sigma_iteration.get("sigma_chol"),
                build_sigma_iteration.get("sigma_chol_inv"),
            ]
            for build_sigma_iteration in build_sigma_map
        ]

        sigma_raw = [sigmas[0] for sigmas in list_sigmas]
        sigma_chol = [sigmas[1] for sigmas in list_sigmas]
        sigma_chol_inv = [sigmas[2] for sigmas in list_sigmas]

        return (sigma_raw, sigma_chol, sigma_chol_inv)

    def __parser_sigma_derivatives(self, sigma_derivatives_map):
        list_sigmas = [
            [build_sigma_iteration.get("sigma_derivative")]
            for build_sigma_iteration in sigma_derivatives_map
        ]
        sigma_derivative = [sigmas[0] for sigmas in list_sigmas]
        return sigma_derivative

    def __parser_c_inverse(self, raw_c_components):

        sigma_chol_block_matrix = raw_c_components[0]
        sigma_chol_inv_block_matrix = raw_c_components[1]
        c_inverse = raw_c_components[2]

        return (sigma_chol_block_matrix, sigma_chol_inv_block_matrix, c_inverse)



import numpy as np

from numpy.linalg import solve
from .mcglmcattr import MCGLMCAttributes


class MCGLMVariance(MCGLMCAttributes):
    """
    The MCGLMVariance class handles the second optimization of the MCGLM second-moment assumptions, therefore, the step for variance. It implements every step of Variance within the scope of the MCGLM algorithm, using many attributes to be specified as attributes. A general class must inherit this MCGLMVariance and leverages its methods properly. MCGLM is in charge of setting the fundamental python attributes and modules orchestration for a complete mcglm adjustment.

    The variance step on the optimization sketch boils down to Pearson estimating equations and the chaser algorithm for optimization. The latter uses tuning to set the step size of each iteration.

    Heavy operations regarding C components, pivotal to the chaser optimization step, are implemented on the MCGLMAttricutes class, inherited here. The method _c_complete, the one that crafts all of the three attributes thoroughly, is comprehensive for the variance calculation in this class.
    """

    def __init__(self):
        super(MCGLMCAttributes, self).__init__()

    def update_covariates(self, mu_attributes, rho, power, tau, W, dispersion, mu):
        """The method update_covariates implements a cycle of iteration for the second-moment estimation, the variance.

        Parameters
        ----------
            mu_attributes : dict
                A dict with mean and derivatives.
            rho : array_type
                Parameters of correlation.
            power : float
                A parameter for Power Tweedie distribution.
            tau : float
                Dispersion parameters.
            W : array_type
                A weight matrix.
            dispersion : array-type
                A vector with dispersion parameters.
            mu : array_type
                A vector with mean parameters.
        Returns
        -------
            tuple: A tuple with new vector of dispersion vector, atributes of matrix C, sensitivity.
        """
        c_inverse, c_derivative, c_values = self.c_complete(
            mu_attributes, power, rho, tau
        )
        (
            pearson_score,
            sensitivity,
            c_intermediate_components,
        ) = self.__pearson_estimating_equation(mu, W, c_inverse, c_derivative)

        step = self.__chaser_step(pearson_score, sensitivity)
        new_covariates = dispersion - step

        return (
            new_covariates,
            c_inverse,
            c_values,
            c_intermediate_components,
            sensitivity,
        )

    def __chaser_step(self, score, sensitivity):
        """The protected method chaser step calculates the step for a optimization step.

        Parameters
        ----------
            score : array_type
                A vector with quasi-score values.
            sensitivity : array_type
                The sensitivity matrix.
        Returns
        -------
            array_type: The absolute change to operate on vector.
        """
        return self._tuning * solve(sensitivity, score)

    def __pearson_estimating_equation(self, mu, W, c_inverse, c_derivative):
        """The estimating equation for the dispersion parameters.

        Parameters
        ----------
            mu : array_type
                A vetor with mean parameters
            W : array_type
                A weight matrix
            c_inverse : array_type
                The inverse of C Matrix
            c_derivative : array_type
                The derivatives of C Matrix
        Returns
        -------
            tuple : a tuple with score, sensitivity matrix and the matrix C normalized by Pearson.
        """
        residue = self._y_values - mu

        c_pearson = [np.dot(c_inverse, d_c) for d_c in c_derivative]

        pearson_score = [
            self.__core_pearson(matrix_component, c_inverse, residue, W)
            for matrix_component in c_pearson
        ]
        sensitivity = self.generate_sensitivity(c_pearson, W)
        return (pearson_score, sensitivity, c_pearson)

    def __core_pearson(self, c_component, c_inverse, residue, W):
        """The protected method core_pearson handles the inner-components of Pearson estimation equations operations.

        Parameters
        ----------
            c_component : array_type
                A matrix with components of C.
            c_inverse : array_type
                The inverse of matrix C.
            residue : array_type
                The difference between mean and an outcome variable.
            W : array_type
                A weight matrix.
        Returns
        -------
            array_type : A matrix with the core pearson.
        """
        weighted_c_components = np.dot(c_component, W)
        sum_diagonal = np.sum(np.diag(weighted_c_components))
        residue_inv_C = np.dot(c_inverse, residue)

        core_pearson = np.subtract(
            np.dot(np.dot(residue.transpose(), weighted_c_components), residue_inv_C),
            sum_diagonal,
        )
        return core_pearson

    @staticmethod
    def generate_sensitivity(c_intermediate_components, W):
        """The method to create the sensitivity matrix.

        Parameters
        ----------
            c_intermediate_components : array_type
                Intermediate components of C matrix.
            W : array_type
                A weight matrix.
        Returns
        -------
            array_type : A sensitivity matrix
        """
        sensitivity = np.array([])

        for position_row in range(len(c_intermediate_components)):
            transpose_matrix_position = (
                c_intermediate_components[position_row].copy().transpose()
            )
            transpose_matrix_position = np.dot(W, transpose_matrix_position)
            for position_col in range(len(c_intermediate_components)):
                sensitivity = np.append(
                    sensitivity,
                    -np.sum(
                        np.multiply(
                            transpose_matrix_position,
                            c_intermediate_components[position_col],
                        )
                    ),
                )
        sensitivity = sensitivity.reshape(
            len(c_intermediate_components), len(c_intermediate_components)
        )
        return sensitivity



import numpy as np
import itertools

from .mcglmcattr import MCGLMCAttributes
from statsmodels.genmod.families.links import (
    Logit,
    Power,
    Log,
    probit,
    cauchy,
    CLogLog,
    LogLog,
    NegativeBinomial,
    inverse_power,
)
from scipy.linalg import block_diag
from numpy.linalg import solve


AVAILABLE_LINK_FUNCTIONS = {
    "logit": Logit(),
    "identity": Power(power=1.0),
    "power": Power(),
    "log": Log(),
    "probit": probit(),
    "cauchy": cauchy(),
    "cloglog": CLogLog(),
    "loglog": LogLog(),
    "negativebinomial": NegativeBinomial(),
    "inverse_power": inverse_power(),
    "reciprocal": Power(),
}


class MCGLMMean(MCGLMCAttributes):
    """
    MCGLMMean is the class for the first moment adjustment within MCGLM inference. It handles lifecycle completely, ranging from mu and derivatives to quasi-likelihood calculations.
    This class has two interfaces: 'calculate_mean_features' which calculates mu attributes, and 'update_beta' that applies quasi-likelihood estimation and retrieves a new beta.
    This class implements the Estimating Equation Quasi-score (Wedderburn, 1974) and the second-order optimization algorithm (Jennrich, 1969) and(Widyaningsih et al., 2017).

    References
    ----------
    Wedderburn, R. W. M. (1974). Quasi-likelihood functions, generalized linear models, and the Gauss—Newton method. Biometrika, 61(3):439–447.

    Jennrich, R. I. (1969). A Newton-Raphson algorithm for maximum likelihood factor analysis. Psychometrika, 34.

    Widyaningsih, P., Saputro, D. e Putri, A. (2017). Fisher scoring method for parameter estimation of geographically weighted ordinal logistic regression (gwolr) model. Journal of Physics: Conference Series, 855:012060.
    """

    def __init__(self):
        super(MCGLMCAttributes, self).__init__()

    def __get_link_function(self, link: str):
        """Check whether the link function is available

        Parameters
        ----------
            link : str
                A link function
        Returns
        -------
            statsmodels.genmod.families.links : a corresponding object for the link function.
        """
        assert link in AVAILABLE_LINK_FUNCTIONS, (
            f"The link function " + str(link) + " isn't available"
        )
        return AVAILABLE_LINK_FUNCTIONS.get(link.lower())

    def __linear_predictor(self, X, beta):
        """Method linear predictor applies linear operation between covariates and the regression parameters.

        Parameters
        ----------
            X : array_like
                Design matrix with covariates
            beta : array-like
                Regression parameters
        Returns
        -------
            array_like : The calculated output vector.
        """
        return np.dot(X, beta)

    def _link_function_attributes(
        self, link: str, beta: np.array, X: np.array, offset: int = 0
    ):
        """A protected method for calling the __link_function_attributes

        Parameters
        ----------
            link : str
                Link function.
            beta : array_like
                Regression Parameters.
            X : array_like
                Matrix with covariates.
            offset : (int, optional)
                Offset add value. Defaults to 0.
        Returns
        -------
            dict : Dict value with the mean and its derivatives.
        """
        return self.__link_function_attributes(link, beta, X, offset)

    def __link_function_attributes(
        self, link: str, beta: np.array, X: np.array, offset: int = 0
    ):
        """The method __link_function_attributes calculates the vector of expected values and its derivatives. It returns data as dictionary

        Parameters
        ----------
            link : str
                Link function.
            beta : array-like
                Regression Parameters.
            X : array-like
                Matrix with covariates.
            offset : int, optional
                Offset add value. Defaults to 0.
        Returns
        -------
            dict : Dict value with the mean and its derivatives.
        """
        link_func = self.__get_link_function(link.lower())
        eta = self.__linear_predictor(X, beta)
        if offset is None:
            offset = 0
        eta = eta + offset

        mu = link_func.inverse(eta)
        deriv = X * link_func.inverse_deriv(eta)[:, None]
        return dict(mu=mu, deriv=deriv)

    def calculate_mean_features(self, link, beta, X, offset):
        """Base method to calculate every attribute related to the mean.

        Parameters
        ----------
            link : str
                Link function.
            beta : array-like
                Regression Parameters.
            X : array-like
                Matrix with covariates.
            offset : int, optional)
                Offset add value. Defaults to 0.
        Returns
        -------
            tuple : Mean attributes, the raw mean and its derivatives.
        """
        mu_attributes_per_response = list(
            map(self.__link_function_attributes, link, beta, X, offset)
        )
        mu = np.array(
            list(
                itertools.chain(
                    *[mu_value.get("mu") for mu_value in mu_attributes_per_response]
                )
            )
        )
        d = block_diag(
            *[mu_value.get("deriv") for mu_value in mu_attributes_per_response]
        )
        return mu_attributes_per_response, mu, d

    def update_beta(self, beta, W, power, rho, tau):
        """The method update_beta takes the current beta, leverages the quasi-likelihood estimator to calculate the next regression parameters.

        Parameters
        ----------
            beta : array-like
                Regression Parameters.
            W : array-like
                Weight matrix
            power : float
                Power parameter
            rho : float
                Correlation parameters
            tau : float
                Dispersion parameters.
        Returns
        -------
            tuple : A tuple with the new regression parameters, the quasi-score parameter, sensitivity and the variability matrix.
        """
        mu_attributes, mu, derivative_mu = self.calculate_mean_features(
            self._link, beta, self._X, self._offset
        )
        c_inverse = self.c_inverse(mu_attributes, power, rho, tau)

        score, sensitivity, variability = self.__quasi_score(
            derivative_mu, c_inverse, self._y_values, mu, W
        )
        new_beta = self.__update_fisher_score(sensitivity, score, beta)

        return new_beta, score, sensitivity, variability

    def __update_fisher_score(self, sensitivity, score, beta):
        """A private method that implements the second-order optimization algorithm Fisher-scoring.

        Parameters
        ----------
            sensitivity : array-like
                Sensitivity matrix
            score : array-like
                Quasi-score output.
            beta : array-like
                Regression parameters.

        Returns
        -------
            array-like : New regression vector.
        """
        linear_system = solve(sensitivity, score)
        index = 0
        adjusted_betas = []
        for beta_value in beta:
            adjusted_betas.append(linear_system[index : index + len(beta_value)])
            index += len(beta_value)
        adjusted_betas = np.array(adjusted_betas)
        new_beta = beta - adjusted_betas

        return new_beta

    def __quasi_score(self, mu_derivative, c_inverse, y, mu, W):
        """Quasi-score functions are estimating equations for the Maximum Likelihood Estimator (Wedderburn, 1974). This method harnesses the Numpy backbone to produce a classic method of statistical models.

        Parameters
        ----------
            mu_derivative : array-like
                First-order derivatives of means.
            c_inverse : array-like
                Inverse of C matrix
            y : array-like
                A vector with outcome variable.
            mu : array-like
                A vetor with mean parameters.
            W : array-like
                A matrix of weights.

        Returns
        -------
            tuple : A tuple with score, sensitivity and variability.
        """

        residue = y - mu

        mu_derivative_transpose = mu_derivative.transpose()
        mu_derivative_and_c = np.dot(mu_derivative_transpose, c_inverse)

        score = np.dot(np.dot(mu_derivative_and_c, W), residue)
        sensitivity = np.dot(np.dot(-mu_derivative_and_c, W), mu_derivative)
        variability = np.dot(np.dot(mu_derivative_and_c, W**2), mu_derivative)

        return (score, sensitivity, variability)



import numpy as np
import statsmodels.api as sm
import pandas as pd


from itertools import groupby
from scipy import stats
from scipy.stats import ncx2
from numpy.linalg import inv, slogdet
from scipy.linalg import block_diag
from statsmodels.iolib.summary import Summary
from statsmodels.genmod.generalized_linear_model import GLM, GLMResults
from statsmodels.genmod.families.links import probit, cauchy, CLogLog, LogLog
from .mcglmmean import MCGLMMean
from .mcglmvariance import MCGLMVariance
from .mcglmcattr import MCGLMCAttributes
from .utils import mc_sandwich


class MCGLM(MCGLMMean, MCGLMVariance):
    __doc__ = """
    MCGLM class that implements MCGLM stastical models. (Bonat, Jørgensen 2015)
        
    It extends GLM for multi-responses and dependent components by fitting second-moment assumptions.

    Parameters
    ----------
    endog : array_like
        1d array of endogenous response variable. In case of multiple responses, the user must pass the responses on a list.  
    exog : array_like
        A dataset with the endogenous matrix in a Numpy fashion. Since the library doesn't set an intercept by default, the user must add it. In the case of multiple responses, the user must pass the design matrices as a python list. 
    z : array_like
        List with matrices components of the linear covariance matrix.
    link : array_like, string or None
        Specification for the link function. The MCGLM library implements the following options: identity, logit, power, log, probit, cauchy, cloglog, loglog, negativebinomial. In the case of None, the library chooses the identity link. In multiple responses, user must pass values as list.  
    variance : array_like, string or None
        Specification for the variance function. The MCGLM library implements the following options: constant, tweedie, binomialP, binomialPQ, geom_tweedie, poisson_tweedie. In the case of None, the library chooses the constant link. In multiple responses, user must pass values as list.   
    offset : array_like or None
        Offset for continuous or count. In multiple responses, user must pass values as list.   
    Ntrial : array_like or None
        Ntrial for binomial responses. In multiple responses, user must pass values as list.
    power_fixed : array_like or None
        Parameter that allows estimation of power when variance functions is either tweedie, geom_tweedie or poisson_tweedie. In multiple responses, user must pass values as list. 
    maxiter : float or None
        Number max of iterations. Defaults to 200.
    tol : float or None
        Threshold of minimum absolute change on paramaters. Defaults to 0.0001.
    tuning : float or None
        Step size parameter. Defaults to 0.5.
    weights : array_like or None
        Weight matrix. Defaults to None.
        
    Examples
    ----------
    >>> import statsmodels.api as sm
    >>> data = sm.datasets.scotland.load()
    >>> data.exog = sm.add_constant(data.exog)
    
    >>> model = sm.GLM(data.endog, data.exog, z=[mc_id(data.exog)],
    ...                      link="log", variance="tweedie",
    ...                      power=2, power_fixed=False)
    
    >>> model_results = model.fit()
    >>> model_results.mu
    >>> model_results.pearson_residuals
    >>> model_results.aic
    >>> model_results.bic
    >>> model_results.loglikelihood
    
    Notes
    -----
    MCGLM is a brand new model, which provides a solid statistical model for fitting multi-responses non-gaussian, dependent, or independent data based on second-moment assumptions. When a user instantiates an mcglm object, she must specify attributes such as link, variance, and z matrices; it will drive the overall behavior of the model.
    For more details, check articles and documentation provided.
    """

    def __init__(
        self,
        endog,
        exog,
        z,
        link=None,
        variance=None,
        offset=None,
        ntrial=None,
        power=None,
        power_fixed=None,
        maxiter=50,
        tol=0.001,
        tuning=1,
        weights=None,
    ):
        super(MCGLM, self).__init__()

        self._max_iter = maxiter
        self._tol = tol
        self._tuning = tuning
        self._weights = weights

        self._n_targets = None
        self._n_obs = None
        self._X = None
        self._y_values = None
        self._y_names = None
        self._z = None
        self._ntrial = None
        self._tau_initial = None
        self._beta_initial = None
        self._rho_initial = None
        self._offset = None
        self._link = None
        self._variance = None
        self._power_fixed = None
        self._power_initial = None

        (
            self._n_targets,
            self._n_obs,
            self._X,
            self._y_values,
            self._y_names,
            self._z,
            self._ntrial,
            self._tau_initial,
            self._beta_initial,
            self._rho_initial,
            self._offset,
            self._link,
            self._variance,
            self._power_fixed,
            self._power_initial,
        ) = self.__calculate_static_attributes(
            endog, exog, link, variance, z, offset, power, power_fixed, ntrial
        )

    @property
    def df_model(self):
        """Calculates the degree of freedom for the model."""
        return len(self._beta_initial[0]) + len(self._tau_initial[0])

    @property
    def df_resid(self):
        """Calculates the degree of freedom for the model residuals."""
        return self._n_obs - self.df_model

    def __calculate_static_attributes(
        self, endog, exog, link, variance, z, offset, power, power_fixed, ntrial
    ):
        """Base method to warm up the pivotal artifacts for model training. It fulfills the None with default values, sets the power value, and creates the base vectors.

        Args:
            endog (list or np.array): _description_
            exog (list or np.array): _description_
            link (list): _description_
            variance (list): _description_
            z (list): _description_
            offset (list): _description_
            power (list): _description_
            power_fixed (list): _description_
            ntrial (list): _description_
        """

        def initial_power(variance):
            if variance in [
                "binomialP",
                "tweedie",
                "geom_tweedie",
                "poisson_tweedie",
                "constant",
            ]:
                return 1
            elif variance == "binomialPQ":
                return (1, 1)
            else:
                return 0

        n_targets = len(exog) if isinstance(exog, list) else 1
        n_obs = exog[0].shape[0] if n_targets > 1 else exog.shape[0]

        set_offset = offset is None
        set_link = link is None
        set_variance = variance is None
        set_power_fixed = power_fixed is None
        set_power = power is None

        tau = list()
        beta = list()
        rho = list()

        offset_list = list()
        link_list = list()
        variance_list = list()
        power_fixed_list = list()
        power_list = list()
        y_values = list()
        y_names = list()

        if n_targets == 1:
            X = [exog]
            y_values = [endog.values]
            y_names = [endog.name]
            z = [z]
            ntrial = [ntrial]
            tau = [[0 for i in range(len(z[0]))]]
            beta = [np.ones(X[0].shape[1])]
            rho = 0

            offset_list = [None] if set_offset else [offset]
            link_list = ["identity"] if set_link else [link]
            variance_list = ["constant"] if set_variance else [variance]
            power_fixed_list = [True] if set_power_fixed else [power_fixed]
            power_list = [initial_power(variance_list[0])] if set_power else [power]

        else:
            X = exog
            if not set_offset:
                offset_list = offset
            if not set_link:
                link_list = link
            if not set_variance:
                variance_list = variance
            if not set_power_fixed:
                power_fixed_list = power_fixed
            if not set_power:
                power_list = power

            for index in range(n_targets):
                tau.append([0 for i in range(len(z[index]))])
                beta.append([np.ones(exog[index].shape[1])])
                y_values.append(endog[index].values)
                y_names.append(endog[index].name)

                if set_offset:
                    offset_list.append(None)
                if set_link:
                    link_list.append("identity")
                if set_variance:
                    variance_list.append("constant")
                if set_power_fixed:
                    power_fixed_list.append(True)
                if set_power:
                    power_list.append(initial_power(variance[index]))

            rho = list()
            for _ in range(int(n_targets * (n_targets - 1) / 2)):
                rho.append(0)

        return (
            n_targets,
            n_obs,
            X,
            np.concatenate(y_values, axis=None),
            y_names,
            z,
            ntrial,
            tau,
            beta,
            rho,
            offset_list,
            link_list,
            variance_list,
            power_fixed_list,
            power_list,
        )

    def fit(self):
        """The interface to run the inference for MCGLM statistical model."""
        (
            regression_historical,
            dispersion_historical,
            residue,
            varcov,
            joint_inv_sensitivity,
            joint_variability,
            n_iter,
            mu,
            rho,
            tau,
            power,
            parameters_target,
            c_inverse,
            c_values,
        ) = self._fit()

        regression_parameters = regression_historical[-1]
        dispersion_parameters = dispersion_historical[-1]

        p_log_likelihood = self.__p_log_likelihood(
            self._y_values, mu, c_values, c_inverse
        )
        aic = self.__p_aic(p_log_likelihood, self.df_model)
        bic = self.__p_bic(p_log_likelihood, self.df_model, len(self._y_values))

        return MCGLMResults(
            normalized_var_cov=abs(varcov),
            nobs=self._n_obs,
            n_targets=self._n_targets,
            y_names=self._y_names,
            regression=regression_parameters,
            dispersion=parameters_target,
            n_iter=n_iter,
            residue=residue,
            rho=rho,
            tau=tau,
            power=power,
            link=self._link,
            variance=self._variance,
            power_fixed=self._power_fixed,
            p_log_likelihood=p_log_likelihood,
            aic=aic,
            bic=bic,
            df_resid=self.df_resid,
            df_model=self.df_model,
            mu=mu.reshape(self._n_targets, -1),
            y_values=self._y_values,
            X=self._X,
            ntrial=self._ntrial,
        )

    def __p_log_likelihood(self, endog, mu, c_values, c_inverse):
        """
        Gaussian Pseudo-loglikelihood
        """

        def gauss(residue, det_sigma, inv_sigma):
            import math

            size_residue = len(residue)
            dens = (
                -1 * (size_residue / 2) * np.log(2 * math.pi)
                - 0.5 * det_sigma
                - (residue.T * inv_sigma * residue)
            )
            return dens.mean()

        residue = endog - mu
        det_sigma = slogdet(block_diag(c_values))[1]
        log_likelihood = gauss(
            residue=residue, det_sigma=det_sigma, inv_sigma=c_inverse
        )

        return log_likelihood

    def __p_aic(self, pseudo_loglike, degrees_of_freedom):
        """
        pseudo Akaike Information criterion
        """
        return round(2 * degrees_of_freedom - 2 * pseudo_loglike, 2)

    def __p_bic(self, pseudo_loglike, degrees_of_freedom, length_endogenous):
        """
        pseudo Bayesian Information criterion
        """
        return round(
            degrees_of_freedom * np.log(length_endogenous) - 2 * pseudo_loglike, 2
        )

    def _fit(self):
        """This method implements the core inference for MCGLM, by all the means of the two-moment assumptions."""
        W = (
            np.diag(np.ones(len(self._y_values)))
            if self._weights is None
            else self._weights
        )

        regression_historical = []
        dispersion_historical = []

        # To a warm up the start
        self.__update_optimal_dispersion()

        rho = self._rho_initial
        tau = self._tau_initial
        power = self._power_initial

        # Setting up
        regression = np.array(self._beta_initial)
        dispersion = np.array(self.__create_dispersion_vector(rho, power, tau))

        regression_historical.append(regression)
        dispersion_historical.append(dispersion)
        for iter in range(self._max_iter):
            # First moment

            (
                new_regression,
                quasi_score,
                mean_sensitivity,
                mean_variability,
            ) = self.update_beta(regression, W, power, rho, tau)
            regression_historical.append(new_regression)

            # Second moment
            mu_attributes, mu, mu_derivatives = self.calculate_mean_features(
                self._link, new_regression, self._X, self._offset
            )

            (
                new_dispersion,
                c_inverse,
                c_values,
                c_derivatives_componentes,
                var_sensitivity,
            ) = self.update_covariates(
                mu_attributes, rho, power, tau, W, dispersion, mu
            )

            dispersion_historical.append(new_dispersion)

            regression, dispersion, rho, power, tau = self.__iteration_update(
                new_regression, new_dispersion, rho, power, tau
            )
            if self.__check_stop_criterion(
                regression_historical, dispersion_historical
            ):
                break

        residue = self._y_values - mu

        (
            varcov,
            joint_inv_sensitivity,
            joint_variability,
        ) = MCGLMParameters._parameters_attributes(
            residue,
            mu_derivatives,
            W,
            quasi_score,
            mean_sensitivity,
            mean_variability,
            c_inverse,
            c_values,
            c_derivatives_componentes,
            var_sensitivity,
        )

        parameters_target = self.__get_dispersion_parameters(mu, rho, tau, dispersion)

        return (
            regression_historical,
            dispersion_historical,
            residue,
            varcov,
            joint_inv_sensitivity,
            joint_variability,
            str(iter),
            mu,
            rho,
            tau,
            power,
            parameters_target,
            c_inverse,
            c_values,
        )

    def __get_dispersion_parameters(self, mu, rho, tau, dispersion):

        covariances = dispersion.tolist()

        if self._n_targets == 1:
            rho = np.array([rho])
        elif self._n_targets > 1:
            rho = np.array([covariances.pop(0) for _ in range(len(rho))])

        parameters_target = list()
        for index in range(self._n_targets):
            power_mu = {
                "power": self._power_initial[index]
                if self._power_fixed[index]
                else covariances.pop(0),
                "mu": mu[index * self._n_obs : (index + 1) * self._n_obs],
            }

            size_parameters = (
                len(tau[index]) if self._power_fixed[index] else len(tau[index])
            )

            scale_output = {
                "scalelist": [
                    round(covariances.pop(0), 3) for _ in range(size_parameters)
                ]
            }

            parameters_target.append({**scale_output, **power_mu})

        return parameters_target

    def __check_stop_criterion(self, regression_historical, dispersion_historical):
        regression_lift = abs(regression_historical[-2] - regression_historical[-1])
        dispersion_lift = abs(dispersion_historical[-2] - dispersion_historical[-1])

        for position in range(len(regression_lift)):
            if np.all(regression_lift[position] < self._tol) and np.all(
                dispersion_lift < self._tol
            ):
                return True
        return False

    def __iteration_update(self, new_regression, new_dispersion, rho, power, tau):

        if self._n_targets == 1:
            if self._power_fixed[0]:
                return new_regression, new_dispersion, rho, power, [new_dispersion]
            else:
                return new_regression, new_dispersion, rho, [new_dispersion[0]], [new_dispersion[1:].tolist()]

        if (isinstance(rho, list)) or (isinstance(rho, np.ndarray)):
            new_rho = new_dispersion[0 : len(rho)]
            stack_index = len(rho)
        else:
            new_rho = new_dispersion[0]
            stack_index = 1

        new_power = list()
        new_tau = list()
        for position in range(len(tau)):
            # power section
            if self._power_fixed[position]:
                new_power.append(power[position])
            else:
                new_power.append(new_dispersion[stack_index])
                stack_index += 1

            # tau section.
            additions = 0
            temp_tau = list()
            for index in range(len(tau[position])):
                temp_tau.append(new_dispersion[stack_index + index])
                additions += 1
            stack_index += additions
            new_tau.append(temp_tau)
            
        return new_regression, new_dispersion, new_rho, new_power, new_tau

    def __create_dispersion_vector(self, rho, power, tau):
        covariance = list()

        if self._n_targets > 1:
            if isinstance(rho, list):
                covariance.extend(rho)
            else:
                covariance.append(rho)
        for n_outputs in range(len(tau)):
            if not self._power_fixed[n_outputs]:
                covariance.append(power[n_outputs])
            covariance = covariance + tau[n_outputs]
        return covariance

    def __update_optimal_dispersion(self):
        """Update optimal dispersion method calculates optimal values for regression parameters and the dispersion. It harnesses the GLM API of statsmodels for calculating those values."""

        def logit_est(endog, exog, offset):
            """For the Logit link function, the method adjusts a GLM with Binomial family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(endog, exog, family=sm.families.Binomial(), offset=offset)
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def loglog_est(endog, exog, offset):
            """For the LogLog link function, the method adjusts a GLM with Binomial family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(endog, exog, family=sm.families.Binomial(LogLog()), offset=offset)
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def cloglog_est(endog, exog, offset):
            """For the CLogLog link function, the method adjusts a GLM with Binomial family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(
                endog, exog, family=sm.families.Binomial(CLogLog()), offset=offset
            )
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def cauchy_est(endog, exog, offset):
            """For the Cauchy link function, the method adjusts a GLM with Binomial family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(endog, exog, family=sm.families.Binomial(cauchy()), offset=offset)
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def probit_est(endog, exog, offset):
            """For the Cauchy link function, the method adjusts a GLM with Binomial family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(endog, exog, family=sm.families.Binomial(probit()), offset=offset)
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def identity_est(endog, exog, offset=None):
            """For the Identity link function, the method adjusts a OLS. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = sm.OLS(endog, exog, offset=offset)
            mdl_results = mdl.fit()
            return mdl_results.params, mdl_results.scale

        def log_est(endog, exog, offset):
            """For the Log link function, the method adjusts a GLM with Tweedie(power=1) family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(
                endog, exog, family=sm.families.Tweedie(var_power=1), offset=offset
            )
            mdl_results = mdl.fit()
            return mdl_results.params, mdl_results.scale

        def power_est(endog, exog, offset):
            """For either the Power or Reciprocal link function, the method adjusts a GLM with Tweedie(power=2) family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(
                endog, exog, family=sm.families.Tweedie(var_power=2), offset=offset
            )
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        def negative_binomial_est(endog, exog, offset):
            """For the Negative Binomial link function, the method adjusts a GLM with Tweedie(power=2) family. It retrieves the parameters specified.

            Args:
                endog (array-like): outcome variable.
                exog (array-like): independent variables.
                offset (int): shift on response variable.

            Returns:
                tuple: regression parameters, dispersion parameter.
            """
            mdl = GLM(
                endog, exog, family=sm.families.Tweedie(var_power=2), offset=offset
            )
            mdl_results = mdl.fit()
            return mdl_results.params.values, mdl_results.scale

        first_estimation = {
            "logit": logit_est,
            "identity": identity_est,
            "power": power_est,
            "log": log_est,
            "probit": probit_est,
            "cauchy": cauchy_est,
            "cloglog": cloglog_est,
            "loglog": loglog_est,
            "negativebinomial": negative_binomial_est,
            "inverse_power": log_est,
            "reciprocal": power_est,
        }

        for target in range(self._n_targets):

            self._beta_initial[target], dispersion = first_estimation[
                self._link[target]
            ](
                self._y_values[
                    target * self._n_obs : target * self._n_obs + self._n_obs
                ],
                self._X[target],
                self._offset[target],
            )
            self._tau_initial[target][0] = dispersion


class MCGLMParameters:
    """According to MCGLM specification, grounded for frequentist inference traits, the estimation of resulting parameters converge asymptotically to a gaussian distribution with tuple mean-variance = (actual parameters, inverse of matrix Godambe). This property allows the calculation of pivotal traits regarding the parameters, such as: hypothesis testing and confidence interval.

    This class implements every method related to this trait.
    """

    @staticmethod
    def _parameters_attributes(
        resid,
        mu_deriv,
        W,
        quasi_score,
        mean_sens,
        mean_varia,
        c_inv,
        c,
        c_deriv,
        var_sensi,
    ):
        """The parameters of MCGLM converge assymtoptically to a Normal Distribution. This trait allows some statistical inferences as hypothesis testing, confidence interval and so on. This method crafts three important matrices: Varcov: variance covariance matrix, joint_inv_sensitivity: inverse sensitivity, and joint_variability: the joint variability distribution."""

        variance_variability = MCGLMParameters.generate_var_variability(
            resid, W, c_inv, c, c_deriv
        )

        inv_cw = np.dot(c_inv, W)

        s_cov_beta = MCGLMParameters._mc_cross_sensitivity(
            cov_product=c_deriv,
            columns_size=len(quasi_score),
        )

        v_cov_beta = MCGLMParameters._mc_cross_variability(
            cov_product=c_deriv,
            inv_cw=inv_cw,
            res=resid,
            d=mu_deriv,
        )

        p1 = np.append(mean_varia, v_cov_beta.transpose(), axis=0)

        p2 = np.append(v_cov_beta, variance_variability, axis=0)
        joint_variability = np.append(p1, p2, axis=1)

        inv_J_beta = inv(mean_sens)
        inv_S_beta = inv_J_beta
        inv_S_cov = inv(var_sensi)
        mat0 = np.zeros((s_cov_beta.shape[1], s_cov_beta.shape[0]))

        cross_term = mc_sandwich(s_cov_beta, -inv_S_cov, inv_S_beta)
        p1 = np.append(inv_S_beta, cross_term, axis=0)
        p2 = np.append(mat0, inv_S_cov, axis=0)

        joint_inv_sensitivity = np.append(p1, p2, axis=1)
        varcov = mc_sandwich(
            joint_variability, joint_inv_sensitivity, joint_inv_sensitivity.transpose()
        )

        return (
            varcov,
            joint_inv_sensitivity,
            joint_variability,
        )

    def _mc_cross_sensitivity(cov_product, columns_size):
        nrow = len(cov_product)
        return np.zeros((nrow, columns_size))

    def generate_var_variability(res, w, c_inv, c_val, c_comp):

        return MCGLMParameters._calculate_variability(
            product=c_comp,
            inv_C=c_inv,
            C=c_val,
            res=res,
            W=w,
        )

    def _calculate_variability(product, inv_C, C, res, W):
        n_par = len(product)
        we = [np.dot(product[index], inv_C) for index in range(n_par)]

        k4 = res**4 - 3 * np.diag(C) ** 2
        sensitivity = MCGLMVariance.generate_sensitivity(product, W=W**2)
        W = np.diag(W).flatten()

        variability = MCGLMParameters._mc_variability(sensitivity, we, k4, W)

        return variability

    def _mc_variability(sensitivity, we, k4, w):
        variability = np.array([])
        for position_row in range(len(we)):
            wi = np.diag(we[position_row])
            for position_col in range(len(we)):
                wj = np.diag(we[position_col])

                k4_operation = np.sum(k4 * wi * w * wj * w)

                variability = np.append(
                    variability,
                    -2 * sensitivity.item((position_row, position_col)) + k4_operation,
                )

        return variability.reshape((len(we), len(we)))

    def _covprod(a, w, res):
        calculation_sandwich = np.dot(np.dot(res, w), res)
        calculation_residue = np.dot(res.transpose(), a)
        product = np.dot(calculation_sandwich, calculation_residue)
        return product

    def _mc_cross_variability(cov_product, inv_cw, res, d):

        wlist = [np.dot(cov, inv_cw) for cov in cov_product]
        a = np.dot(d.transpose(), inv_cw)

        n_beta = a.shape[0]
        n_cov = len(cov_product)
        cross_variability = []
        for cov in range(n_cov):
            for beta in range(n_beta):
                cross_variability.append(
                    MCGLMParameters._covprod(a[beta, :], wlist[cov], res)
                )
        cross_variability = np.array(cross_variability).reshape(n_cov, n_beta).T
        return cross_variability


class MCGLMResults(GLMResults):
    """MCGLM Class for generating and manipulating results of mcglm training. The main output goes by the method summary(), the classical statsmodels output. Therefore, the user can access the attributes "aic", "bic" e loglikelihood.

    Args:
        GLMResults: Class of statsmodels library for presenting results of GLM.
    """

    def __init__(
        self,
        normalized_var_cov,
        nobs,
        n_targets,
        y_names,
        regression,
        dispersion,
        n_iter,
        residue,
        rho,
        tau,
        power,
        link,
        variance,
        power_fixed,
        p_log_likelihood,
        aic,
        bic,
        df_resid,
        df_model,
        mu,
        y_values,
        X,
        ntrial,
    ):
        self._normalized_var_cov = normalized_var_cov
        self._nobs = nobs
        self._n_targets = n_targets
        self._y_names = y_names
        self._regression = regression
        self._dispersion = dispersion
        self._n_iter = n_iter
        self._residue = residue
        self._rho = rho
        self._tau = tau
        self._power_list = power
        self._link_list = link
        self._variance_list = variance
        self._power_fixed_list = power_fixed
        self._p_log_likelihood = p_log_likelihood
        self._p_aic = aic
        self._p_bic = bic
        self._df_resid = df_resid
        self._df_model = df_model
        self._mu_value = mu
        self._y_values = y_values
        self._X = X
        self._ntrial = ntrial
        self.params = None
        #self.bse = None

        self._use_t = False
        self.model = None

    @property
    def aic(self):
        return self._p_aic

    @property
    def bic(self):
        return self._p_bic

    @property
    def loglikelihood(self):
        return self._p_log_likelihood

    @property
    def bse(self):
        """The standard errors of the parameter estimates."""
        if (not hasattr(self, "cov_params_default")) and (
            self.normalized_cov_params is None
        ):
            bse_ = np.empty(len(self.params))
            bse_[:] = np.nan
        else:
            bse_ = np.sqrt(np.diag(self.normalized_cov_params))
        return bse_

    @property
    def tvalues(self):
        """
        Return the t-statistic for a given parameter estimate.
        """
        return self.params / self.bse

    @property
    def mu(self):
        return self._mu_value

    @property
    def pvalues(self):
        """The two-tailed p values for the t-stats of the params."""
        if self.use_t:
            df_resid = getattr(self, "df_resid_inference", self._df_resid)
            return stats.t.sf(np.abs(self.tvalues), df_resid) * 2
        else:
            return stats.norm.sf(np.abs(self.tvalues)) * 2

    @property
    def pearson_residuals(self):
        """
        Pearson residuals.  The Pearson residuals are defined as
        (`endog` - `mu`)/sqrt(VAR(`mu`)) where VAR is the distribution
        specific variance function.  See statsmodels.families.family and
        statsmodels.families.varfuncs for more information.
        """
        residuals = []
        if self._n_targets == 1:
            mu = [self.mu]
        else:
            mu = self.mu.reshape(self._n_targets, -1)

        for index in range(self._n_targets):
            residue = (
                self._y_values[index * self._nobs : index * self._nobs + self._nobs]
                - mu[index]
            )

            variance = None
            if self._variance_list[index] in ("binomialP", "binomialPQ"):
                variance = self._variance_list[index]
            else:
                variance = "power"

            variance_sqrt_output = (
                MCGLMCAttributes()
                ._generate_variance(
                    variance,
                    mu[index],
                    self._power_list[index],
                    self._ntrial[index],
                )
                .get("variance_sqrt_output")
            )

            residuals.append(residue / np.diag(variance_sqrt_output))

        return residuals

    @property
    def vcov(self):
        return self.normalized_cov_params

    def anova(
        self, indexes_covariates=[[1, 2, 2, 2, 2]], covariate_name=[["x1", "x2"]]
    ):

        total_anovas = list()
        for position_target in range(self.model.n_targets):
            dispersion = self.model.dispersion[position_target]["scalelist"][1:].copy()
            dispersion_vcov = self._dispersion_vcov[position_target][1:, 1:].copy()

            covariates_positions = [
                list(j) for i, j in groupby(indexes_covariates[position_target])
            ]
            covariates_names = covariate_name[position_target]

            index = 0
            index_name = 0
            anovas = list()
            for covs in covariates_positions:

                current_cov = np.array(dispersion[index : index + len(covs)])
                current_vcov = dispersion_vcov[
                    index : index + len(covs), index : index + len(covs)
                ]

                try:
                    solve_vcoc = inv(current_vcov)
                except Exception as e:
                    solve_vcoc = current_vcov

                chi_square = np.dot(
                    current_cov.transpose(), np.dot(solve_vcoc, current_cov)
                )
                df = len(covs)
                pvalue = ncx2.pdf(chi_square, df, nc=0)

                index += len(covs)
                anovas.append(
                    {
                        "covariate_name": covariates_names[index_name],
                        "chi_square": round(chi_square, 4),
                        "df": df,
                        "pvalue": round(pvalue, 4),
                    }
                )

                index_name += 1
            total_anovas.append(anovas)

        return total_anovas

    def __add_table_two_columns(
        self, summ, yname="output", xname=None, title=None, alpha=0.05
    ):
        top_left = [
            ("Dep. Variable:", None),
            ("Model:", ["MCGLM"]),
            ("link:", [self.link]),
            ("variance:", [self.variance]),
            ("Method:", ["Quasi-Likelihood"]),
            ("Date:", None),
            ("Time:", None),
        ]

        top_right = [
            ("No. Iterations:", [self._n_iter]),
            ("No. Observations:", [self._nobs]),
            ("Df Residuals:", [self._df_resid]),
            ("Df Model:", [self._df_model]),
            ("Power-fixed:", [self.power_fixed]),
            ("pAIC", [self.paic]),
            ("pBIC", [self.pbic]),
            ("pLogLik", [round(self._p_log_likelihood, 4)]),
        ]

        if hasattr(self, "cov_type"):
            top_left.append(("Covariance Type:", [self.cov_type]))

        if title is None:
            title = "Multivariate Covariance Generalized Linear Model"

        summ.add_table_2cols(
            self,
            gleft=top_left,
            gright=top_right,
            yname=yname,
            xname=xname,
            title=title,
        )
        summ.add_table_params(
            self, yname=yname, xname=xname, alpha=alpha, use_t=self.use_t
        )

        return summ

    def __add_dispersion(self, summ, title=None, alpha=0.05):
        summ.add_table_params(
            self,
            alpha=alpha,
            xname=["dispersion_" + str(i + 1) for i in range(len(self.params))],
            use_t=self.use_t,
        )

        return summ

    def __add_power(self, summ, alpha=0.05):
        summ.add_table_params(
            self,
            alpha=alpha,
            xname=[
                "power_" + (str(i + 1) if len(self.params) > 1 else "")
                for i in range(len(self.params))
            ],
            use_t=self.use_t,
        )

        return summ

    def __add_rho_section(self, summ, alpha=0.05):
        self.normalized_cov_params = self._rho_vcov
        self.params = np.array(self._rho) if isinstance(self._rho, int) else self._rho
        if isinstance(self._rho, (np.ndarray, np.generic)):
            summ.add_table_params(
                self,
                alpha=alpha,
                xname=[
                    "rho_" + (str(i + 1) if len(self.params) > 1 else "")
                    for i in range(len(self.params))
                ],
                use_t=self.use_t,
            )

            if hasattr(self, "constraints"):
                summ.add_extra_txt(
                    [
                        "Model has been estimated subject to linear "
                        "equality constraints."
                    ]
                )
        return summ

    def summary(self, yname=None, xname=None, title=None, alpha=0.05):
        """
        It generates the summary report as the sketch of classical "statsmodels" library. The summary shows all parameters found thoroughly, for each response.
        """
        self._dispersion_vcov = []
        self.betas_vcov = []
        self._rho_vcov = []
        self.power_vcov = []

        index_position = 0
        cov_params = self._normalized_var_cov

        if self._n_targets == 1:
            self._rho_vcov = np.array([np.nan])
            mu = [self.mu]
        else:
            mu = self.mu

        for index in range(self._n_targets):

            n_betas = len(self._regression[index])
            self.betas_vcov.append(
                cov_params[
                    index_position : index_position + n_betas,
                    index_position : index_position + n_betas,
                ]
            )
            index_position += n_betas

        if self._n_targets > 1:
            self._rho_vcov = cov_params[
                index_position : index_position + len(self._rho),
                index_position : index_position + len(self._rho),
            ]

            index_position += len(self._rho)

        for index in range(self._n_targets):

            if self._power_fixed_list[index]:
                self.power_vcov.append(np.array([np.nan]))
            else:
                self.power_vcov.append(
                    np.array(
                        cov_params[
                            index_position : index_position + 1,
                            index_position : index_position + 1,
                        ]
                    )
                )
                index_position += 1

            n_dispersion = len(self._dispersion[index]["scalelist"])

            self._dispersion_vcov.append(
                cov_params[
                    index_position : index_position + n_dispersion,
                    index_position : index_position + n_dispersion,
                ]
            )

            index_position += n_dispersion

        self.scale = 1

        smry = Summary()

        for target_index in range(self._n_targets):
            exog_names = self._X[target_index].columns.tolist()
            y_name = self._y_names[target_index]

            self.link = self._link_list[target_index]
            self.variance = self._variance_list[target_index]
            self.power_fixed = self._power_fixed_list[target_index]
            self.params = self._regression[target_index]
            self.normalized_cov_params = self.betas_vcov[target_index]

            self.loglikehood = self._p_log_likelihood
            self.paic = self._p_aic
            self.pbic = self._p_bic

            smry = self.__add_table_two_columns(smry, yname=y_name, xname=exog_names)

            self.params = np.array(self._dispersion[target_index]["scalelist"])
            self.normalized_cov_params = self._dispersion_vcov[target_index]

            #self.bse = np.sqrt(np.diag(self.normalized_cov_params))
            smry = self.__add_dispersion(smry)
            self.params = np.array([self._dispersion[target_index]["power"]])
            self.normalized_cov_params = self.power_vcov[target_index]

            smry = self.__add_power(smry)

        smry = self.__add_rho_section(smry)

        return smry



__version__ = "0.1.0"

from mcglm.dependencies import mc_id, mc_ma, mc_mixed
from mcglm.mcglm import MCGLM, MCGLMResults
from mcglm.mcglmcattr import MCGLMCAttributes
from mcglm.mcglmmean import MCGLMMean
from mcglm.mcglmvariance import MCGLMVariance

__all__ = [
    "mc_id",
    "mc_ma",
    "mc_mixed",
    "MCGLM",
    "MCGLMCAttributes",
    "MCGLMMean",
    "MCGLMVariance",
]



import numpy as np
from functools import reduce


def diagonal(n: int, values: np.array):
    raw_matrix = np.zeros((n, n), float)
    np.fill_diagonal(raw_matrix, list(values))
    return raw_matrix


def mc_sandwich(central_matrix, left_matrix, right_matrix):
    return np.dot(np.dot(left_matrix, central_matrix), right_matrix)


def mc_sandwich_power(central_matrix, left_matrix, right_matrix):
    matrix_operation = mc_sandwich(central_matrix, left_matrix, right_matrix)
    return matrix_operation + matrix_operation.transpose()


def mc_matrix_linear_predictor(tau: list, z: list):
    map_operation = map(lambda x, y: x * y, tau, z)
    liner_predictor_calculus = reduce(lambda x, y: x + y, map_operation)
    return liner_predictor_calculus


def mc_sandwich_csr(central_matrix, left_matrix, right_matrix):
    return left_matrix.dot(central_matrix).dot(right_matrix)


def mc_sandwich_power_csr(central_matrix, left_matrix, right_matrix):
    matrix_operation = mc_sandwich_csr(central_matrix, left_matrix, right_matrix)
    return matrix_operation + matrix_operation.transpose()



"""
An extension of MCGLM library to provide three options for matrix linear predictor: mc_id, mc_ma, and mc_mixed.
"""
import numpy as np
import itertools

from patsy import dmatrix
from mcglm.utils import diagonal
from itertools import combinations


def mc_id(data=None):
    """
    mc_id method retrieves a numpy diagonal matrix with data length of the original matrix
    """
    size = data.shape[0]
    return diagonal(size, np.ones(size))


def mc_ma(id=None, time=None, data=None, order=1):
    """
    mc_ma method retrieves the Z components for matrix linear predictor associated with Autoregressive models(Feller, W. (1957). An introduction to probability theory and its applications / William Feller.
    Wiley New York, 2nd ed. edition.)
    """

    def diagonal_position(indexes=None, k=1):
        new_indexes = tuple((indexes[0][:-k] + k, indexes[1][:-k]))
        return new_indexes

    try:
        data = data.sort_values(by=[id, time], ascending=True)
    except Exception as e:
        print("Parameter data must be a pandas DataFrame")
        raise e

    min_time = sorted(np.unique(data[time]))[:order]
    id = data[id].values
    time = data[time].values

    sample_size = len(id)
    ma = np.zeros((sample_size, sample_size))

    ma_vector = np.ones(sample_size - order)
    index_time = sorted(
        list(itertools.chain(*[np.where(time == value)[0] for value in min_time]))
    )

    for index in index_time[order:]:
        ma_vector[index - order] = 0

    indexes_pos = np.diag_indices(sample_size)
    indexes = diagonal_position(indexes_pos, order)

    ma[indexes] = ma_vector

    return ma + ma.T


def mc_mixed(data=None, formula=None):
    """
    mc_mixed retrieves the components for matrix linear predictor associated with mixed models(Demidenko E (2013). Mixed Models: Theory and Applications with R. John Wiley & Sons. doi:10.1002/0471728438.).
    """
    design_matrix = dmatrix(formula, data=data, return_type="dataframe")
    val_columns = design_matrix.columns.tolist()

    # Find two-points
    positions = list()
    for column in val_columns:
        if ":" in column:
            positions.append(1)
        else:
            positions.append(0)
    positions = np.array(positions)

    design_matrix = design_matrix.values

    all_indexes = [(i, i) for i in set(positions)]

    if len(all_indexes) > 1:
        all_indexes = all_indexes + list(itertools.combinations(list(range(2)), 2))

    matrices = list()
    for tc in all_indexes:
        # matrix1 = np.repeat(design_matrix[:, tc[0]], size, axis=0).reshape(size, size).T
        # matrix2 = np.repeat(design_matrix[:, tc[1]], size, axis=0).reshape(size, size).T
        matrix1 = design_matrix[:, np.where(positions == tc[0])[0]]
        matrix2 = design_matrix[:, np.where(positions == tc[1])[0]]

        if tc[0] == tc[1]:
            matrices.append(np.matmul(matrix1, matrix2.T))
        else:
            matrices.append(
                np.matmul(matrix1, matrix2.T) + np.matmul(matrix2, matrix1.T)
            )

    return matrices



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcglm']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.2,<4.0.0',
 'numpy==1.22',
 'pandas>=1.3.3,<2.0.0',
 'patsy>=0.5.2,<0.6.0',
 'scipy==1.7.1',
 'seaborn>=0.11.2,<0.12.0',
 'statsmodels>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'mcglm',
    'version': '0.2.1',
    'description': 'Multivariate Covariance Generalized Linear Models',
    'long_description': '### Multivariate Covariance Generalized Linear Models\n\nhttps://pypi.org/project/mcglm/\n\nThe mcglm package brings to python language one of the most powerful extensions to GLMs(Nelder, Wedderburn; 1972), the Multivariate Covariance Generalized Linear Models(Bonat, Jørgensen; 2016).\n\nThe GLMs have consolidated as a unified statistical model for analyzing non-gaussian independent data throughout the years. Notwithstanding enhancements to Linear Regression Models(Gauss), some key assumptions, such as the independence of components in the response, each element of the target belonging to an exponential dispersion family maintains.\n\nMCGLM aims to expand the GLMs by allowing fitting on a wide variety of inner-dependent datasets, such as spatial and longitudinal, and supplant the exponential dispersion family output by second-moment assumptions(Wedderburn; 1974)\n\nhttps://jeancmaia.github.io/posts/tutorial-mcglm/tutorial_mcglm.html\n\n-----\n\nThe mcglm python package follows the standard pattern of the statsmodels library and aims to be another API on the package. Therefore, Python machine learning practitioners will be very familiar with this new statistical model. \n\n\nTo install this package, use \n\n```bash\npip install mcglm\n```\n',
    'author': 'Jean Carlos Maia',
    'author_email': 'jeanclmaia@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<=3.9',
}


setup(**setup_kwargs)
